/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "liberty.h"
#include "patterns.h"

static int mouse_shapes(int *i, int *j, int *vala, char *patname_a,
			int *k, int *l, int *valb, 
			char *patname_b, int color,
			int goal[MAX_BOARD][MAX_BOARD], struct pattern *type);
static void mouse_shapes_callback(int m, int n, int color,
				  struct pattern *pattern, int ll, void *data);
static int mouse_determine_life(int goal[MAX_BOARD][MAX_BOARD],
				struct eye_data eye[MAX_BOARD][MAX_BOARD]);

static int shapes_ai;     /* Location, value and pattern for best move found so far */
static int shapes_aj;
static int shapes_vala;
static char shapes_name_a[10]; 
static int shapes_bi;     /* Second best move found so far */
static int shapes_bj;
static int shapes_valb;
static char shapes_name_b[10]; 

#define UNUSED(x)  x=x


/* Returns true if a move can be found to attack the dragon
 * at (m,n), in which case (*i,*j) is the recommended move.
 *
 * The array goal marks the extent of the dragon. This must
 * be maintained during reading.
 */

int
mouse_attack (int m, int n, int *ui, int *uj, int goal[MAX_BOARD][MAX_BOARD])
{
  int vala, valb, i, j, k, l;
  char patname_a[10], patname_b[10];
  int true_genus;
  
  if (stackp==0)
    for (i=0; i<board_size; i++)
      for (j=0; j<board_size; j++) {
	if ((dragon[i][j].origini==dragon[m][n].origini)
	    && (dragon[i][j].originj==dragon[m][n].originj))
	  goal[i][j]=1;
	else
	  goal[i][j]=0;
      }
  
  if (mouse_shapes(&i, &j, &vala, patname_a, &k, &l, &valb, patname_b,
		   OTHER_COLOR(p[m][n]), goal, apat)) {
    if ((vala>30) && stackp <= depth) {
      if ((i != -1) && trymove(i, j, OTHER_COLOR(p[m][n]), patname_a, m, n)) {
	int save_goal=goal[i][j];
	if (!mouse_defend(m, n, NULL, NULL, goal)) {
	  int save_sgf_dump=sgf_dump;
	  int save_count_variations=count_variations;
	  sgf_dump=0;
	  count_variations=0;
	  if (p[m][n]==BLACK)
	    true_genus=mouse_determine_life(goal, black_eye);
	  else 
	    true_genus=mouse_determine_life(goal, white_eye);
	  sgf_dump=save_sgf_dump;
	  count_variations = save_count_variations;
	  /* It is now the defender's move so we lose if he has 1.5 eyes. */
	  if (true_genus < 3) {
	    popgo();
	    goal[i][j]=save_goal;
	    if (ui) *ui=i;
	    if (uj) *uj=j;
	    return 1;
	  }
	}
	popgo();
	goal[i][j]=save_goal;
      }
      if ((k != -1) && trymove(k, l, OTHER_COLOR(p[m][n]), patname_b, m, n)) {
	int save_goal=goal[k][l];
	if (!mouse_defend(m, n, NULL, NULL, goal)) {
	  int save_sgf_dump=sgf_dump;
	  int save_count_variations=count_variations;
	  sgf_dump=0;
	  count_variations=0;
	  if (p[m][n]==BLACK)
	    true_genus=mouse_determine_life(goal, black_eye);
	  else 
	    true_genus=mouse_determine_life(goal, white_eye);
	  sgf_dump=save_sgf_dump;
	  count_variations = save_count_variations;
	  /* It is now the defender's move so we lose if he has 1.5 eyes. */
	  if (true_genus < 3) {
	    popgo();
	    goal[k][l]=save_goal;
	    if (ui) *ui=k;
	    if (uj) *uj=l;
	    return 1;
	  }
	}
	popgo();
	goal[k][l]=save_goal;
      }
    }
  }
  return 0;
}
      


/* Returns true if a move can be found to defend the dragon
 * at (m,n), in which case (*i,*j) is the recommended move.
 *
 * The array goal marks the extent of the dragon. This must
 * be maintained during reading.  */

int
mouse_defend (int m, int n, int *ui, int *uj, int goal[MAX_BOARD][MAX_BOARD])
{
  int vala, valb, i, j, k, l;
  char patname_a[10], patname_b[10];
  int true_genus=0;
  
  if (stackp==0)
    for (i=0; i<board_size; i++)
      for (j=0; j<board_size; j++) {
	if ((dragon[i][j].origini==dragon[m][n].origini)
	    && (dragon[i][j].originj==dragon[m][n].originj))
	  goal[i][j]=1;
	else
	  goal[i][j]=0;
      }
  
  if (mouse_shapes(&i, &j, &vala, patname_a, &k, &l, &valb, patname_b,
		   p[m][n], goal, dpat)) {
    if ((vala>30) && stackp <= depth) {
      if ((i != -1) && trymove(i, j, p[m][n], patname_a, m, n)) {
	int save_goal=goal[i][j];
	goal[i][j]=1;
	if (!mouse_attack(m, n, NULL, NULL, goal)) {
	  int save_sgf_dump=sgf_dump;
	  int save_count_variations=count_variations;
	  sgf_dump=0;
	  count_variations=0;
	  if (p[m][n]==BLACK)
	    true_genus=mouse_determine_life(goal, black_eye);
	  else 
	    true_genus=mouse_determine_life(goal, white_eye);
	  sgf_dump=save_sgf_dump;
	  count_variations = save_count_variations;
	  /* It is now the attacker's move so we need two full eyes */
	  if (true_genus >= 4) {
	    popgo();
	    goal[i][j]=save_goal;
	    if (ui) *ui=i;
	    if (uj) *uj=j;
	    return 1;
	  }
	}
	popgo();
	goal[i][j]=save_goal;
      }
      if ((k != -1) && trymove(k, l, p[m][n], patname_b, m, n)) {
	int save_goal=goal[k][l];
	goal[k][l]=1;
	if (!mouse_attack(m, n, NULL, NULL, goal)) {
	  int save_sgf_dump=sgf_dump;
	  int save_count_variations=count_variations;
	  sgf_dump=0;
	  count_variations=0;
	  if (p[m][n]==BLACK)
	    true_genus=mouse_determine_life(goal, black_eye);
	  else 
	    true_genus=mouse_determine_life(goal, white_eye);
	  sgf_dump=save_sgf_dump;
	  count_variations = save_count_variations;
	  /* It is now the attacker's move so we need two full eyes */
	  if (true_genus >= 4) {
	    popgo();
	    goal[k][l]=save_goal;
	    if (ui) *ui=k;
	    if (uj) *uj=l;
	    return 1;
	  }
	}
	popgo();
	goal[k][l]=save_goal;
      }
    }
  }
  return 0;
}


/* The support of goal is a string for which no defensive move is
 * found. If the string is BLACK then eye is black_eye, and if
 * the string is WHITE then eye is white_eye. The function runs make_domains
 * and computes the number of eyes. Returns twice the number of half eyes.
 */

static int
mouse_determine_life(int goal[MAX_BOARD][MAX_BOARD], struct eye_data eye[MAX_BOARD][MAX_BOARD])
{
  int mw[MAX_BOARD][MAX_BOARD];
  int true_genus = 0;
  int max, min;
  int attacki, attackj;
  int m, n;

  memset(mw, 0, sizeof(mw));
  make_domains();
  /* The eyespaces we want to evaluate are the one which
   * are adjacent to the dragon (whose stones comprise the
   * support of goal) which are not GRAY_BORDERED. These
   * are the eyespaces of the dragon. */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      if (p[m][n] && goal[m][n]) {
	if ((m>0) && (p[m-1][n] == EMPTY) && 
	    eye[m-1][n].color != GRAY_BORDER)
	  mw[eye[m-1][n].origini][eye[m-1][n].originj]=1;
	if ((m<board_size-1) && (p[m+1][n] == EMPTY) && 
	    eye[m+1][n].color != GRAY_BORDER)
	  mw[eye[m+1][n].origini][eye[m+1][n].originj]=1;
	if ((n>0) && (p[m][n-1] == EMPTY) &&
	    eye[m][n-1].color != GRAY_BORDER)
	  mw[eye[m][n-1].origini][eye[m][n-1].originj]=1;
	if ((n<board_size-1) && (p[m][n+1] == EMPTY) &&
	    eye[m][n+1].color != GRAY_BORDER)
	  mw[eye[m][n+1].origini][eye[m][n+1].originj]=1;
      }

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) 
      if (mw[m][n]) {
      compute_eyes(m, n, &max, &min, &attacki, &attackj, eye);
      true_genus += (max+min);
      /* found two eyes---look no further */
      if (true_genus >= 4)
	return 4;
    }
  return true_genus;
}

	
/* Generates one or two moves, attempting to attack or defend the dragon at
 * (goali, goalj). The function sets (*i, *j) to the first move found, or 
 * (-1, -1) if no move is found. The second move (*k, *l) is handled
 * similarly. The values of the moves found are left in *val1 and *val2.
 *
 * Returns number of moves found, or 0 if no move is found.
 */



/* This function takes (*i, *j) to be either (-1,-1) or the
 * location of an already-found move with value *vala, and
 * similarly an optional second move (*k, *l) with value
 * valb. It looks for moves to replace these. Only moves
 * near (goali, goalj) are considered.
 */

static int 
mouse_shapes(int *i, int *j, int *vala, char patname_a[10],
	     int *k, int *l, int *valb, char patname_b[10], 
	     int color, int goal[MAX_BOARD][MAX_BOARD], struct pattern *type)
{
  int m, n; /* iterate over all board posns */
  int save_sgf_dump=sgf_dump;
  int save_count_variations=count_variations;
  sgf_dump=0;
  count_variations=0;

  *i = -1;
  *j = -1;
  *vala = -1;
  *k = -1;
  *l = -1;
  *valb = -1;

  shapes_vala = -1;
  shapes_valb = -1;
  shapes_ai = -1;
  shapes_aj = -1;
  shapes_bi = -1;
  shapes_bj = -1;

  compile_for_match();
  for (m = 0; m < board_size; m++)
    for (n = 0; n < board_size; n++)
      if (p[m][n])
	matchpat(m, n, mouse_shapes_callback, color, type, NULL, goal);

  *i=shapes_ai;
  *j=shapes_aj;
  *vala=shapes_vala;
  strncpy(patname_a, shapes_name_a, 10);
  *k=shapes_bi;
  *l=shapes_bj;
  *valb=shapes_valb;
  strncpy(patname_b, shapes_name_b,10);

  sgf_dump=save_sgf_dump;
  count_variations=save_count_variations;
  if (*valb > 0)
    return 2;
  else if (*vala > 0)
    return 1;
  else
    return 0;
}  


static void
mouse_shapes_callback(int m, int n, int color, struct pattern *pattern,
		      int ll, void *data)
{
  int ti, tj, tval;  /* trial move and its value */

  /* Don't need auxiliary data in this callback. */
  UNUSED(data);
  
  /* Pick up the location of the move */
  TRANSFORM(pattern->movei, pattern->movej, &ti, &tj, ll);
  ti += m;
  tj += n;

  /* For sacrifice patterns, the survival of the stone to be played is
   * not checked (but it still needs to be legal). Otherwise we
   * discard moves which can be captured. 
   */
  if (! (pattern->class & CLASS_s)) {
    if (!safe_move(ti, tj, color)) {
      TRACE("  move at %m wasn't safe, discarded\n", ti, tj);
      return;
    }
  }
  else {
    if (! legal(ti, tj, color)) {
      TRACE("  move at %m wasn't legal, discarded\n", ti, tj);
      return;
    }
  }
  
  /* For class n patterns, the pattern is contingent on an opponent
   * move at * not being captured.
   */
  if (pattern->class & CLASS_n) {
    if (!safe_move(ti, tj, OTHER_COLOR(color))) {
      TRACE("  opponent can't play safely at %m, move discarded\n", ti, tj);
      return;
    }
  }
  
  /* If the pattern has a constraint, call the autohelper to see
   * if the pattern must be rejected.
   */
  if (pattern->autohelper_flag & HAVE_CONSTRAINT) {
    if (!pattern->autohelper(pattern, ll, ti, tj, color, 0))
      return;
  }

  /* and work out the value of this move */
  if (pattern->helper) {
    /* ask helper function to consider the move */
    DEBUG(DEBUG_HELPER, "  asking helper to consider '%s'+%d at %m\n", 
	  pattern->name, ll, ti,tj);
    tval = pattern->helper(pattern, ll, ti, tj, color);
    
    if (tval > 0) {
      TRACE("helper likes pattern '%s' value %d at %m\n",
	    pattern->name, tval, ti, tj);
    }
    else {
      DEBUG(DEBUG_HELPER,"  helper does not like pattern '%s' at %m\n", 
	    pattern->name, ti, tj);
      return;  /* pattern matcher does not like it */
    }
  }
  else { /* no helper */
    tval = pattern->value;
  }

  /* having made it here, we have made it through all the extra checks */

  TRACE("Pattern %s found at %m\n", pattern->name, ti, tj);
  if (tval > shapes_vala) {
    if ((ti != shapes_ai) || (tj != shapes_aj)) {
      shapes_valb=shapes_vala;
      shapes_bi=shapes_ai;
      shapes_bj=shapes_aj;
      strncpy(shapes_name_a, pattern->name, 10);
      shapes_ai=ti;
      shapes_aj=tj;
    }
    shapes_vala=tval;
  }
  else if ((tval > shapes_valb) && ((ti != shapes_ai) || (tj != shapes_aj))) {
    shapes_valb=tval;
    shapes_bi=ti;
    shapes_bj=tj;
    strncpy(shapes_name_b, pattern->name, 10);
  }
  assert ((shapes_bi == -1) || (shapes_bi != shapes_ai) || (shapes_bj != shapes_aj));
}


  
/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
